# TPK4186 - Advanced Tools for Performance Engineering
# AssemblyLine.py
# by Antoine Rauzy
# Copyrights (c) 2019-2020 NTNU
#
# This file is part of the educational material provided to NTNU students registered to the course TPK4186.
# It is licensed under a Creative Commons Attribution-NonCommercial 4.0 International License.
# For details about this license, see https://creativecommons.org/licenses/by-nc/4.0/.

# Table of Contents
# -----------------
# 1. Required modules
# 2. LineState
# 3. Line
# 4. Main

# 1. Required modules
# -------------------

import sys

sys.path.insert(0, '..')
import Alea.DiscreteEventSystems
import Alea.DiscreteEventSimulators

# 2. LineState
# ------------

class LineState(Alea.DiscreteEventSystems.State):
  WEIGHT_PART1 = 5
  WEIGHT_PART2 = 2
  MAXIMUM_WEIGHT = 20
  
  def __init__(self, line):
    Alea.DiscreteEventSystems.State.__init__(self, line)
    self.numberOfPart1InBuffer = 0
    self.numberOfPart2InBuffer = 0
    self.weightOfPartsInBuffer = 0
    self.numberOfFinalProducts = 0

  @staticmethod
  def New(line):
    return LineState(line)

  def Reset(self):
    self.numberOfPart1InBuffer = 0
    self.numberOfPart2InBuffer = 0
    self.weightOfPartsInBuffer = 0
    self.numberOfFinalProducts = 0

  def CanPart1BeProduced(self):
    return True

  def ProducePart1(self):
    self.numberOfPart1InBuffer += 1

  def CanPart2BeProduced(self):
    return True

  def ProducePart2(self):
    self.numberOfPart2InBuffer += 1

  def CanPartsBeAssembled(self):
    return self.numberOfPart1InBuffer>=1 and self.numberOfPart2InBuffer>=2

  def AssembleParts(self):
    self.numberOfPart1InBuffer -= 1
    self.numberOfPart2InBuffer -= 2
    self.numberOfFinalProducts += 1

  def IsBufferOverweighted(self):
    return self.numberOfPart1InBuffer*LineState.WEIGHT_PART1 + self.numberOfPart2InBuffer*LineState.WEIGHT_PART2 > LineState.MAXIMUM_WEIGHT

  def GetNumberOfFinalProducts(self):
    return self.numberOfFinalProducts

# 3. Line
# -------

line = Alea.DiscreteEventSystems.DiscreteEventSystem(LineState.New)
T1 = line.NewTransition("ProducePart1", LineState.CanPart1BeProduced, LineState.ProducePart1, Alea.DiscreteEventSystems.UniformDelay(10, 12))
T2 = line.NewTransition("ProducePart2", LineState.CanPart2BeProduced, LineState.ProducePart2, Alea.DiscreteEventSystems.UniformDelay(3, 4))
T3 = line.NewTransition("AssembleParts", LineState.CanPartsBeAssembled, LineState.AssembleParts, Alea.DiscreteEventSystems.UniformDelay(8, 11))

# 4. Main
# -------

simulator = Alea.DiscreteEventSimulators.DiscreteEventSimulator()
simulator.SetSystem(line)
I1 = simulator.NewRewardMeanValue("I1", LineState.GetNumberOfFinalProducts, False)
I2 = simulator.NewPredicateSatisfactionTime("I2", LineState.IsBufferOverweighted, False)

simulator.MonteCarloSimulation(600, 10000)

print("I1\t{0:.2f}".format(I1.GetMean()))
print("I2\t{0:.2f}".format(I2.GetMean()))
