# TPK4186 - Advanced Tools for Performance Engineering
# HelicopterEngine.py
# by Antoine Rauzy
# Copyrights (c) 2019-2020 NTNU
#
# This file is part of the educational material provided to NTNU students registered to the course TPK4186.
# It is licensed under a Creative Commons Attribution-NonCommercial 4.0 International License.
# For details about this license, see https://creativecommons.org/licenses/by-nc/4.0/.

# Table of Contents
# -----------------
# 1. Required modules
# 2. HelicopterState
# 3. Line
# 4. Main

# 1. Required modules
# -------------------

import sys

sys.path.insert(0, '..')
import Alea.DiscreteEventSystems
import Alea.DiscreteEventSimulators

# 2. HelicopterState
# ------------------

class HelicopterState(Alea.DiscreteEventSystems.State):
  
  def __init__(self, helicopter):
    Alea.DiscreteEventSystems.State.__init__(self, helicopter)
    self.isLeftFrontFastenerOK = True
    self.isRightFrontFastenerOK = True
    self.isLeftRearFastenerOK = True
    self.isRightRearFastenerOK = True
    self.numberOfFastenersOK = 4
    self.isLanded = False

  @staticmethod
  def New(helicopter):
    return HelicopterState(helicopter)

  def Reset(self):
    self.isLeftFrontFastenerOK = True
    self.isRightFrontFastenerOK = True
    self.isLeftRearFastenerOK = True
    self.isRightRearFastenerOK = True
    self.numberOfFastenersOK = 4
    self.isLanded = False


  def CanLeftFrontFastenerBreakFirst(self):
    return not self.isLanded and self.isLeftFrontFastenerOK and self.numberOfFastenersOK==4

  def CanLeftFrontFastenerBreakSecond(self):
    return not self.isLanded and self.isLeftFrontFastenerOK and self.numberOfFastenersOK==3

  def CanLeftFrontFastenerBreakThird(self):
    return not self.isLanded and self.isLeftFrontFastenerOK and self.isRightRearFastenerOK and self.numberOfFastenersOK==2

  def CanLeftFrontFastenerBreakEventually(self):
    return not self.isLanded and self.isLeftFrontFastenerOK and ((not self.isRightRearFastenerOK and self.numberOfFastenersOK==2) or self.numberOfFastenersOK<2)

  def BreakLeftFrontFastener(self):
    self.isLeftFrontFastenerOK = False
    self.numberOfFastenersOK -= 1


  def CanRightFrontFastenerBreakFirst(self):
    return not self.isLanded and self.isRightFrontFastenerOK and self.numberOfFastenersOK==4

  def CanRightFrontFastenerBreakSecond(self):
    return not self.isLanded and self.isRightFrontFastenerOK and self.numberOfFastenersOK==3

  def CanRightFrontFastenerBreakThird(self):
    return not self.isLanded and self.isRightFrontFastenerOK and self.isLeftRearFastenerOK and self.numberOfFastenersOK==2

  def CanRightFrontFastenerBreakEventually(self):
    return not self.isLanded and self.isRightFrontFastenerOK and ((not self.isLeftRearFastenerOK and self.numberOfFastenersOK==2) or self.numberOfFastenersOK<2)

  def BreakRightFrontFastener(self):
    self.isRightFrontFastenerOK = False
    self.numberOfFastenersOK -= 1


  def CanLeftRearFastenerBreakFirst(self):
    return not self.isLanded and self.isLeftRearFastenerOK and self.numberOfFastenersOK==4

  def CanLeftRearFastenerBreakSecond(self):
    return not self.isLanded and self.isLeftRearFastenerOK and self.numberOfFastenersOK==3

  def CanLeftRearFastenerBreakThird(self):
    return not self.isLanded and self.isLeftRearFastenerOK and self.isRightFrontFastenerOK and self.numberOfFastenersOK==2

  def CanLeftRearFastenerBreakEventually(self):
    return not self.isLanded and self.isLeftRearFastenerOK and ((not self.isRightFrontFastenerOK and self.numberOfFastenersOK==2) or self.numberOfFastenersOK<2)

  def BreakLeftRearFastener(self):
    self.isLeftRearFastenerOK = False
    self.numberOfFastenersOK -= 1


  def CanRightRearFastenerBreakFirst(self):
    return not self.isLanded and self.isRightRearFastenerOK and self.numberOfFastenersOK==4

  def CanRightRearFastenerBreakSecond(self):
    return not self.isLanded and self.isRightRearFastenerOK and self.numberOfFastenersOK==3

  def CanRightRearFastenerBreakThird(self):
    return not self.isLanded and self.isRightRearFastenerOK and self.isLeftFrontFastenerOK and self.numberOfFastenersOK==2

  def CanRightRearFastenerBreakEventually(self):
    return not self.isLanded and self.isRightRearFastenerOK and ((not self.isLeftFrontFastenerOK and self.numberOfFastenersOK==2) or self.numberOfFastenersOK<2)

  def BreakRightRearFastener(self):
    self.isRightRearFastenerOK = False
    self.numberOfFastenersOK -= 1


  def IsEmergencyCondition(self):
    return not self.isLanded and self.numberOfFastenersOK<4

  def PerformEmergencyLanding(self):
    self.isLanded = True


  def IsEngineUnfastened(self):
    return not self.isLanded and self.numberOfFastenersOK==0

# 3. Helicopter
# -------------

helicopter = Alea.DiscreteEventSystems.DiscreteEventSystem(HelicopterState.New)

meanTimeToFirstFailure = 120000
meanTimeToSecondFailure = 600
meanTimeToThirdFailure = 10
delayToEmergencyLanding = 120
delayToEventualFailure = 0

LF1 = helicopter.NewTransition("BreakLeftFrontFastenerFirst", HelicopterState.CanLeftFrontFastenerBreakFirst, HelicopterState.BreakLeftFrontFastener, Alea.DiscreteEventSystems.ExponentialDelay(meanTimeToFirstFailure))
LF2 = helicopter.NewTransition("BreakLeftFrontFastenerSecond", HelicopterState.CanLeftFrontFastenerBreakSecond, HelicopterState.BreakLeftFrontFastener, Alea.DiscreteEventSystems.ExponentialDelay(meanTimeToSecondFailure))
LF3 = helicopter.NewTransition("BreakLeftFrontFastenerThird", HelicopterState.CanLeftFrontFastenerBreakThird, HelicopterState.BreakLeftFrontFastener, Alea.DiscreteEventSystems.ExponentialDelay(meanTimeToThirdFailure))
LFE = helicopter.NewTransition("BreakLeftFrontFastenerEventually", HelicopterState.CanLeftFrontFastenerBreakEventually, HelicopterState.BreakLeftFrontFastener, Alea.DiscreteEventSystems.DiracDelay(delayToEventualFailure))

RF1 = helicopter.NewTransition("BreakRightFrontFastenerFirst", HelicopterState.CanRightFrontFastenerBreakFirst, HelicopterState.BreakRightFrontFastener, Alea.DiscreteEventSystems.ExponentialDelay(meanTimeToFirstFailure))
RF2 = helicopter.NewTransition("BreakRightFrontFastenerSecond", HelicopterState.CanRightFrontFastenerBreakSecond, HelicopterState.BreakRightFrontFastener, Alea.DiscreteEventSystems.ExponentialDelay(meanTimeToSecondFailure))
RF3 = helicopter.NewTransition("BreakRightFrontFastenerThird", HelicopterState.CanRightFrontFastenerBreakThird, HelicopterState.BreakRightFrontFastener, Alea.DiscreteEventSystems.ExponentialDelay(meanTimeToThirdFailure))
RFE = helicopter.NewTransition("BreakRightFrontFastenerEventually", HelicopterState.CanRightFrontFastenerBreakEventually, HelicopterState.BreakRightFrontFastener, Alea.DiscreteEventSystems.DiracDelay(delayToEventualFailure))

LR1 = helicopter.NewTransition("BreakLeftRearFastenerFirst", HelicopterState.CanLeftRearFastenerBreakFirst, HelicopterState.BreakLeftRearFastener, Alea.DiscreteEventSystems.ExponentialDelay(meanTimeToFirstFailure))
LR2 = helicopter.NewTransition("BreakLeftRearFastenerSecond", HelicopterState.CanLeftRearFastenerBreakSecond, HelicopterState.BreakLeftRearFastener, Alea.DiscreteEventSystems.ExponentialDelay(meanTimeToSecondFailure))
LR3 = helicopter.NewTransition("BreakLeftRearFastenerThird", HelicopterState.CanLeftRearFastenerBreakThird, HelicopterState.BreakLeftRearFastener, Alea.DiscreteEventSystems.ExponentialDelay(meanTimeToThirdFailure))
LRE = helicopter.NewTransition("BreakLeftRearFastenerEventually", HelicopterState.CanLeftRearFastenerBreakEventually, HelicopterState.BreakLeftRearFastener, Alea.DiscreteEventSystems.DiracDelay(delayToEventualFailure))

RR1 = helicopter.NewTransition("BreakRightRearFastenerFirst", HelicopterState.CanRightRearFastenerBreakFirst, HelicopterState.BreakRightRearFastener, Alea.DiscreteEventSystems.ExponentialDelay(meanTimeToFirstFailure))
RR2 = helicopter.NewTransition("BreakRightRearFastenerSecond", HelicopterState.CanRightRearFastenerBreakSecond, HelicopterState.BreakRightRearFastener, Alea.DiscreteEventSystems.ExponentialDelay(meanTimeToSecondFailure))
RR3 = helicopter.NewTransition("BreakRightRearFastenerThird", HelicopterState.CanRightRearFastenerBreakThird, HelicopterState.BreakRightRearFastener, Alea.DiscreteEventSystems.ExponentialDelay(meanTimeToThirdFailure))
RRE = helicopter.NewTransition("BreakRightRearFastenerEventually", HelicopterState.CanRightRearFastenerBreakEventually, HelicopterState.BreakRightRearFastener, Alea.DiscreteEventSystems.DiracDelay(delayToEventualFailure))

EL = helicopter.NewTransition("PerformEmergencyLanding", HelicopterState.IsEmergencyCondition, HelicopterState.PerformEmergencyLanding, Alea.DiscreteEventSystems.DiracDelay(delayToEmergencyLanding))

# 4. Main
# -------

simulator = Alea.DiscreteEventSimulators.DiscreteEventSimulator()
simulator.SetSystem(helicopter)
I1 = simulator.NewPredicateSatisfactionCount("I1", HelicopterState.IsEngineUnfastened, False)
I2 = simulator.NewTransitionFiringCount("I2", EL, False)

simulator.MonteCarloSimulation(600, 1000000)

print("I1\t{0:1.2e}".format(I1.GetMean()))
print("I2\t{0:1.2e}".format(I2.GetMean()))
