# TPK4186 - Advanced Tools for Performance Engineering
# HorseGaits.py
# by Antoine Rauzy
# Copyrights (c) 2019-2020 NTNU
#
# This file is part of the educational material provided to NTNU students registered to the course TPK4186.
# It is licensed under a Creative Commons Attribution-NonCommercial 4.0 International License.
# For details about this license, see https://creativecommons.org/licenses/by-nc/4.0/.

# Table of Contents
# -----------------
# 1. Required modules
# 2. HorseState
# 3. Horses
# 4. Main

# 1. Required modules
# -------------------

import sys

sys.path.insert(0, '..')
import Alea.DiscreteEventSystems
import Alea.DiscreteEventSimulators

# 2. Horse States
# ----------------

class HorseState(Alea.DiscreteEventSystems.State):

  def __init__(self, horse):
    Alea.DiscreteEventSystems.State.__init__(self, horse)
    self.leftFrontLegUp = False
    self.rightFrontLegUp = False
    self.leftHindLegUp = False
    self.rightHindLegUp = False
    self.leftFrontLegTurn = False
    self.rightFrontLegTurn = False
    self.leftHindLegTurn = False
    self.rightHindLegTurn = False

  @staticmethod
  def New(horse):
    return HorseState(horse)

  def Reset(self):
    self.leftFrontLegUp = False
    self.rightFrontLegUp = False
    self.leftHindLegUp = False
    self.rightHindLegUp = False
    if self.system.gait==Horse.WALK:
      self.leftFrontLegTurn = False
      self.rightFrontLegTurn = False
      self.leftHindLegTurn = True
      self.rightHindLegTurn = False
    elif self.system.gait==Horse.TROT:
      self.leftFrontLegTurn = False
      self.rightFrontLegTurn = True
      self.leftHindLegTurn = True
      self.rightHindLegTurn = False

  def IsLeftFrontLegTurn(self):
    return not self.leftFrontLegUp and self.leftFrontLegTurn

  def IsLeftFrontLegUp(self):
    return self.leftFrontLegUp

  def LiftUpLeftFrontLeg(self):
    self.leftFrontLegUp = True

  def PutDownLeftFrontLeg(self):
    self.leftFrontLegUp = False
    if self.system.gait==Horse.WALK:
      self.leftFrontLegTurn = False
      self.rightHindLegTurn = True
    elif self.system.gait==Horse.TROT:
      self.leftFrontLegTurn = False
      self.rightFrontLegTurn = True

  def IsRightFrontLegTurn(self):
    return not self.rightFrontLegUp and self.rightFrontLegTurn

  def IsRightFrontLegUp(self):
    return self.rightFrontLegUp

  def LiftUpRightFrontLeg(self):
    self.rightFrontLegUp = True

  def PutDownRightFrontLeg(self):
    self.rightFrontLegUp = False
    if self.system.gait==Horse.WALK:
      self.rightFrontLegTurn = False
      self.leftHindLegTurn = True
    elif self.system.gait==Horse.TROT:
      self.leftFrontLegTurn = True
      self.rightFrontLegTurn = False

  def IsLeftHindLegTurn(self):
    return not self.leftHindLegUp and self.leftHindLegTurn

  def IsLeftHindLegUp(self):
    return self.leftHindLegUp

  def LiftUpLeftHindLeg(self):
    self.leftHindLegUp = True

  def PutDownLeftHindLeg(self):
    self.leftHindLegUp = False
    if self.system.gait==Horse.WALK:
      self.leftFrontLegTurn = True
      self.leftHindLegTurn = False
    elif self.system.gait==Horse.TROT:
      self.leftHindLegTurn = False
      self.rightHindLegTurn = True

  def IsRightHindLegTurn(self):
    return not self.rightHindLegUp and self.rightHindLegTurn

  def IsRightHindLegUp(self):
    return self.rightHindLegUp

  def LiftUpRightHindLeg(self):
    self.rightHindLegUp = True

  def PutDownRightHindLeg(self):
    self.rightHindLegUp = False
    if self.system.gait==Horse.WALK:
      self.rightFrontLegTurn = True
      self.rightHindLegTurn = False
    elif self.system.gait==Horse.TROT:
      self.leftHindLegTurn = True
      self.rightHindLegTurn = False

  def GetNumberOfHorseStateUp(self):
    count = 0
    if self.leftFrontLegUp:
      count += 1
    if self.rightFrontLegUp:
      count += 1
    if self.leftHindLegUp:
      count += 1
    if self.rightHindLegUp:
      count += 1
    return count

# 3. Horses
# ---------

class Horse(Alea.DiscreteEventSystems.DiscreteEventSystem):
  WALK = 1
  TROT = 2
  GALLOP = 3

  def __init__(self):
    Alea.DiscreteEventSystems.DiscreteEventSystem.__init__(self, HorseState.New)
    self.NewTransition("LiftUpLeftFrontLeg", HorseState.IsLeftFrontLegTurn, HorseState.LiftUpLeftFrontLeg, Alea.DiscreteEventSystems.DiracDelay(0))
    self.NewTransition("PutDownLeftFrontLeg", HorseState.IsLeftFrontLegUp, HorseState.PutDownLeftFrontLeg, Alea.DiscreteEventSystems.DiracDelay(1))
    self.NewTransition("LiftUpRightFrontLeg", HorseState.IsRightFrontLegTurn, HorseState.LiftUpRightFrontLeg, Alea.DiscreteEventSystems.DiracDelay(0))
    self.NewTransition("PutDownRightFrontLeg", HorseState.IsRightFrontLegUp, HorseState.PutDownRightFrontLeg, Alea.DiscreteEventSystems.DiracDelay(1))
    self.NewTransition("LiftUpLeftHindLeg", HorseState.IsLeftHindLegTurn, HorseState.LiftUpLeftHindLeg, Alea.DiscreteEventSystems.DiracDelay(0))
    self.NewTransition("PutDownLeftHindLeg", HorseState.IsLeftHindLegUp, HorseState.PutDownLeftHindLeg, Alea.DiscreteEventSystems.DiracDelay(1))
    self.NewTransition("LiftUpRightHindLeg", HorseState.IsRightHindLegTurn, HorseState.LiftUpRightHindLeg, Alea.DiscreteEventSystems.DiracDelay(0))
    self.NewTransition("PutDownRightHindLeg", HorseState.IsRightHindLegUp, HorseState.PutDownRightHindLeg, Alea.DiscreteEventSystems.DiracDelay(1))
    self.gait = Horse.WALK

# 4. Main
# -------

horse = Horse()
horse.gait = Horse.TROT

simulator = Alea.DiscreteEventSimulators.DiscreteEventSimulator()
simulator.SetSystem(horse)
I1 = simulator.NewRewardMaximumValue("I1", HorseState.GetNumberOfHorseStateUp, False)
I2 = simulator.NewPredicateSatisfactionCount("I2", HorseState.IsLeftFrontLegUp, False)
I3 = simulator.NewPredicateSatisfactionCount("I3", HorseState.IsRightFrontLegUp, False)
I4 = simulator.NewPredicateSatisfactionCount("I4", HorseState.IsLeftHindLegUp, False)
I5 = simulator.NewPredicateSatisfactionCount("I5", HorseState.IsRightHindLegUp, False)

simulator.MonteCarloSimulation(31, 1)

print("I1\t{0:.2f}".format(I1.GetMaximumValue()))
print("I2\t{0:.2f}".format(I2.GetMean()))
print("I3\t{0:.2f}".format(I3.GetMean()))
print("I4\t{0:.2f}".format(I4.GetMean()))
print("I5\t{0:.2f}".format(I5.GetMean()))

